@@ -3,7 +3,7 @@
 """A wrapper for Paramiko that attempts to make ssh sessions easier to work with.  It also contains the
 shell_command function for running local shell scripts!
 """
-# Copyright (C) 2015-2017 Jesse Almanrode
+# Copyright (C) 2015-2018 Jesse Almanrode
 #
 #     This program is free software: you can redistribute it and/or modify
 #     it under the terms of the GNU Lesser General Public License as published by
@@ -20,29 +20,30 @@ shell_command function for running local shell scripts!
 from __future__ import print_function
 from collections import namedtuple, OrderedDict
 from getpass import getuser
-from subprocess import Popen, PIPE, STDOUT
 import logging
 import os
 import paramiko
 import socket
-import warnings
 
+# Privates
 __author__ = 'Jesse Almanrode (jesse@almanrode.com)'
 
-ShellCommand = namedtuple('ShellCommand', ['cmd', 'stdout', 'stderr', 'return_code'])
-ShellCommandCombined = namedtuple('ShellCommandCombined', ['cmd', 'stdout', 'return_code'])
+# Globals
+Command = namedtuple('Command', ['cmd', 'stdout', 'stderr', 'return_code'])
 
 
 def envvars():
-    """ Attempt to determine the current username and location of any ssh keys.  If any value is unable to be determined
-    it is returned as 'None'.
+    """ Attempt to determine the current username and location of any ssh private keys.
+    If any value is unable to be determined it is returned as 'None'.
 
-    :return: NamedTuple of (username, rsa_key, dsa_key, ecdsa_key)
+    This method also checks for any private keys loaded into the SSH Agent.
+
+    :return: NamedTuple of (username, agent_keys, rsa_key, dsa_key, ecdsa_key)
     """
-    env = OrderedDict(user=None, rsa_key=None, dsa_key=None, ecdsa_key=None)
+    env = OrderedDict(username=None, agent_keys=None, rsa_key=None, dsa_key=None, ecdsa_key=None)
     EnvVars = namedtuple('EnvVars', env.keys())
     if os.getlogin() == getuser():
-        env['user'] = getuser()
+        env['username'] = getuser()
     userhome = os.path.expanduser('~')
     if os.path.exists(userhome + "/.ssh"):
         keyfiles = os.listdir(userhome + "/.ssh")
@@ -52,76 +53,37 @@ def envvars():
             env['dsa_key'] = userhome + "/.ssh/id_dsa"
         if 'id_ecdsa' in keyfiles:
             env['ecdsa_key'] = userhome + '/.ssh/id_ecdsa'
+    env['agent_keys'] = paramiko.Agent().get_keys()
     return EnvVars(**env)
 
 
-def shell_command(command, combine=False, decodebytes=True):
-    """Run a command in the shell on localhost and return the output
-
-    :param command: String containing the shell script to run
-    :param combine: Direct stderr to stdout
-    :param decodebytes: Decode bytes objects to unicode strings
-    :return: NamedTuple for (cmd, stdout, stderr) or (cmd, stdout)
-    """
-    if combine:
-        pipeout = Popen(command, shell=True, stdout=PIPE, stderr=STDOUT)
-        stdout, stderr = pipeout.communicate()
-        assert stderr is None
-        if decodebytes:
-            result = ShellCommandCombined(cmd=command, stdout=stdout.decode().strip(),
-                                          return_code=pipeout.returncode)
-        else:
-            result = ShellCommandCombined(cmd=command, stdout=stdout.strip(), return_code=pipeout.returncode)
-    else:
-        pipeout = Popen(command, shell=True, stdout=PIPE, stderr=PIPE)
-        stdout, stderr = pipeout.communicate()
-        if decodebytes:
-            result = ShellCommand(cmd=command, stdout=stdout.decode().strip(),
-                                  stderr=stderr.decode().strip(), return_code=pipeout.returncode)
-        else:
-            result = ShellCommand(cmd=command, stdout=stdout.strip(), stderr=stderr.strip(),
-                                  return_code=pipeout.returncode)
-    return result
-
-
-def do_shell_script(command, combine=False):
-    """ Alias to shell_command.
-
-    .. warning::
-
-        This call will be removed in v4.0
-
-    :param command: String containing the shell script to run
-    :param combine: Combine stderr and stdout in output
-    :return: NamedTuple for (cmd, stdout, stderr) or (cmd, stdout)
-    """
-    warnings.warn('<do_shell_script> will be replaced by <shell_command> in v4.0')
-    return shell_command(command, combine=combine)
-
-
 class SSH(object):
     """SSH Session object
 
     :param fqdn: Fully qualified domain name or IP address
     :param username: SSH username
     :param password: SSH password
-    :param keyfile: SSH keyfile (can be used instead of password)
+    :param keyfile: SSH private key file
+    :param keypass: SSH private key password
     :param port: SSH port
     :param timeout: SSH connection timeout in seconds
     :param connect: Initiate the connect
     :return: SSH connection object
     :raises: SSHException
     """
-    def __init__(self, fqdn, username=None, password=None, keyfile=None, port=22, timeout=30, connect=True):
-        if keyfile is None and username is None:
-            raise paramiko.SSHException('You must specify a password or keyfile')
+    def __init__(self, fqdn, username=None, password=None, keyfile=None, keypass=None, port=22, timeout=30, connect=True):
+        if not keyfile:
+            if len(paramiko.Agent().get_keys()) == 0:
+                if not all((username, password)):
+                    paramiko.SSHException('username and password or keyfile not provided')
         self.host = fqdn
         self.username = username
         self.password = password
-        if keyfile is not None:
+        if keyfile:
             self.keyfile = os.path.abspath(os.path.expanduser(keyfile))
         else:
             self.keyfile = keyfile
+        self.keypass = keypass
         self.port = port
         self.timeout = timeout
         self._connection = paramiko.SSHClient()
@@ -129,8 +91,11 @@ class SSH(object):
         if connect:
             self.__connect()
 
+    def __str__(self):
+        return self.__dict__
+
     def __enter__(self):
-        if self.__is_alive() is False:
+        if self.__alive() is False:
             self.__connect()
         return self
 
@@ -144,14 +109,13 @@ class SSH(object):
         :param dstfile: Path to the remote file
         :return: Result of paramiko.SFTPClient.put()
         """
-        if self.__is_alive() is False:
-            raise paramiko.SSHException("Connection is not established")
+        if self.__alive() is False:
+            raise paramiko.SSHException("connection to %s not established" % (self.host,))
         sftp = paramiko.SFTPClient.from_transport(self._connection.get_transport())
         try:
             result = sftp.put(os.path.expanduser(srcfile), os.path.expanduser(dstfile), confirm=True)
-        except IOError as err:
+        finally:
             sftp.close()
-            raise err
         return result
 
     def sftp_get(self, srcfile, dstfile):
@@ -161,15 +125,13 @@ class SSH(object):
         :param dstfile: Path to the local file
         :return: Result of paramiko.SFTPClient.get()
         """
-        if self.__is_alive() is False:
-            raise paramiko.SSHException("Connection is not established")
+        if self.__alive() is False:
+            raise paramiko.SSHException("connection to %s not established" % (self.host,))
         sftp = paramiko.SFTPClient.from_transport(self._connection.get_transport())
         try:
             result = sftp.get(os.path.expanduser(srcfile), os.path.expanduser(dstfile))
-        except IOError as err:
+        finally:
             sftp.close()
-            raise err
-        sftp.close()
         return result
 
     def ssh_command(self, command, timeout=30, combine=False, decodebytes=True):
@@ -182,31 +144,30 @@ class SSH(object):
         :return: Namedtuple of (cmd, stdout, stderr, return_code) or (cmd, stdout, return_code)
         :raises: SSHException
         """
-        if self.__is_alive() is False:
-            raise paramiko.SSHException("Connection is not established")
+        if self.__alive() is False:
+            raise paramiko.SSHException("connection to %s not established" % (self.host, ))
         if combine:
             try:
                 stdin, stdout, stderr = self._connection.exec_command(command, timeout=timeout, get_pty=True)
                 if decodebytes:
-                    result = ShellCommandCombined(cmd=command, stdout=stdout.read().decode().strip(),
-                                                  return_code=stdout.channel.recv_exit_status())
+                    result = Command(cmd=command, stdout=stdout.read().decode().strip(), stderr=None,
+                                     return_code=stdout.channel.recv_exit_status())
                 else:
-                    result = ShellCommandCombined(cmd=command, stdout=stdout.read().strip(),
-                                                  return_code=stdout.channel.recv_exit_status())
+                    result = Command(cmd=command, stdout=stdout.read().strip(), stderr=None,
+                                     return_code=stdout.channel.recv_exit_status())
             except (paramiko.buffered_pipe.PipeTimeout, socket.timeout):
-                result = ShellCommandCombined(cmd=command, stdout='Command timed out', return_code=124)
+                result = Command(cmd=command, stdout='command timed out', stderr=None, return_code=124)
         else:
             try:
                 stdin, stdout, stderr = self._connection.exec_command(command, timeout=timeout)
                 if decodebytes:
-                    result = ShellCommand(cmd=command, stdout=stdout.read().decode().strip(),
-                                          stderr=stderr.read().decode().strip(),
-                                          return_code=stdout.channel.recv_exit_status())
+                    result = Command(cmd=command, stdout=stdout.read().decode().strip(),
+                                     stderr=stderr.read().decode().strip(), return_code=stdout.channel.recv_exit_status())
                 else:
-                    result = ShellCommand(cmd=command, stdout=stdout.read().strip(), stderr=stderr.read().strip(),
-                                          return_code=stdout.channel.recv_exit_status())
+                    result = Command(cmd=command, stdout=stdout.read().strip(), stderr=stderr.read().strip(),
+                                     return_code=stdout.channel.recv_exit_status())
             except (paramiko.buffered_pipe.PipeTimeout, socket.timeout):
-                result = ShellCommand(cmd=command, stdout='', stderr='Command timed out', return_code=124)
+                result = Command(cmd=command, stdout='', stderr='command timed out', return_code=124)
         return result
 
     def close(self):
@@ -214,10 +175,9 @@ class SSH(object):
 
         :return: None
         """
-        self._connection.close()
-        return None
+        return self._connection.close()
 
-    def is_alive(self):
+    def alive(self):
         """Is an SSH connection alive
 
         :return: True or False
@@ -229,38 +189,45 @@ class SSH(object):
             if self._connection.get_transport().is_alive():
                 return True
             else:
-                raise paramiko.SSHException("Unable to determine state of ssh session")
+                raise paramiko.SSHException("unable to determine state of ssh connection")
 
     def reconnect(self):
         """Alias to connect
         """
         return self.__connect()
 
-    def connect(self):
+    def connect(self, failfast=True):
         """Opens an SSH Connection
 
         :return: True
         :raises: SSHException
         """
+        if self.__alive():
+            raise paramiko.SSHException("connection to % already established" % (self.host, ))
+        # Fail Fast for unreachable host:port
+        if failfast:
+            try:
+                s = socket.socket()
+                s.settimeout(1)
+                s.connect((self.host, self.port))
+            except socket.timeout:
+                raise paramiko.SSHException('connect to host %s port %d: Operation timed out' % (self.host, self.port))
+            else:
+                s.close()
         paramiko.util.logging.getLogger().setLevel(logging.CRITICAL)  # Keeping paramiko from logging errors to stdout
-        if self.__is_alive():
-            raise paramiko.SSHException("Connection is already established")
-        if self.keyfile is not None:
-            if self.username is not None:  # Key file with a custom username!
-                self._connection.connect(self.host, port=self.port, username=self.username,
-                                         key_filename=self.keyfile, timeout=self.timeout, look_for_keys=False)
-            else:
-                self._connection.connect(self.host, port=self.port, key_filename=self.keyfile,
-                                         timeout=self.timeout, look_for_keys=False)
-        else:  # Username and password combo
-            if self.username is None or self.password is None:
-                raise paramiko.SSHException("You must provide a username and password or supply an SSH key")
-            else:
-                self._connection.connect(self.host, port=self.port, username=self.username,
-                                         password=self.password, timeout=self.timeout, look_for_keys=False)
+        if not self.keyfile:
+            if len(paramiko.Agent().get_keys()) == 0:
+                if not all((self.username, self.password)):
+                    paramiko.SSHException('username and password or keyfile not provided')
+        if self.keyfile:
+            self._connection.connect(self.host, port=self.port, username=self.username, password=self.keypass,
+                                     key_filename=self.keyfile, timeout=self.timeout, look_for_keys=False)
+        else:
+            self._connection.connect(self.host, port=self.port, username=self.username, password=self.password,
+                                     timeout=self.timeout, look_for_keys=False)
         return True
 
     # Privatizing some of the functions so SSH can be subclassed
-    __is_alive = is_alive
+    __alive = alive
     __connect = connect
     __close = close
